      program expnfit

C NON-LINEAR LEAST-SQUARES FIT OF ONE TO FOUR EXPONENTIAL DECAY
C CALL JPL ROUTINES; LINK AGAINST THE LIB OR COPY ALL ROUTINES 
C ADDED OPTIONAL PGPLOT SUPPORT; DEPENDS ON ENV VAR PGPLOT_DEV

      integer    i, j, k, pgopen, npre
      character*80 str,ttl,tmp

      integer  mxpts, mxcol, npts, nexp, icol
      parameter (mxpts=500000, mxcol=12)
      real*8  r(mxpts), rmsr, c0, pts(mxcol)
      real*4  time(mxpts), corr(mxpts), theo(mxpts)
      real*4  x1,x2, y1,y2, c

      integer   liv, lwk, mxcf, ncf, ifixt, iplat, ntau, iolev
      parameter (mxcf=8)
      parameter (liv=82+4*mxcf)
      parameter (lwk=105+mxcf*(mxpts+2*mxcf+21)+2*mxpts)
      integer iv(liv)
      real*8  wrk(lwk), cf(mxcf), bnds(2,mxcf), a, cfi(mxcf)
      real*8  zero, one, tolr, dranu
      parameter (zero = 0.d0, one = 1.d0, tolr = 1.d-4)
      character*80 fname
      logical qrept

C LABELED LOCATIONS IN ARRAY IV()
      integer outlev, statpr, solprt, covprt, rdreq
      parameter(outlev=19, solprt=22, statpr=23, covprt=14)
      parameter(rdreq=57)

C Z  = INDEPENDENT VARIABLE
C FZ = F(Z), OBSERVED DATA
C TZ = T(Z), THEORETICAL FIT

      real*8  z(mxpts),fz(mxpts), tz(mxpts), plat,tau,ccf
      common /fitdat/ z, fz, tz, plat, tau, ccf, c0

C POSSIBLE MODELS; ROUTINES COMPUTE RESIDUALS BASED ON TARGET FXN
C MUST BE DECLARED EXTERNAL BECAUSE THE NAME IS PASSED AS AN ARGUMENT
      external exp1f
      external exp2f
      external exp3f
      external exp4f
      external exp2ft
      external exp3ft
      external exp4ft
      external exp1fp
      external exp2fp
      external exp3fp
C RANDOM NUMBER ROUTINE, PGPLOT
      external dranu
      external pgopen

C ::::::::::::: READ CONTROL FILE FROM UNIT 3 :::::::::::::::::::::
C LINE1 :: INPUT_DATA_FILE_NAME (assumed form filename.ext)
C LINE2 :: NEXP IFIXT IPLAT IOLEV NCOL ICOL
C LINE3 :: (CF(J),J=1,NEXP-1), [ PLAT ]
C LINE4 :: (TF(J),J=1,NEXP)
C LINE5 :: OUTPUT_DATA_FILE_NAME (optional; otherwise, derived)
C 
C NEXP  :: NO. OF EXPONENTIALS TO USE FOR FIT
C IFIXT :: FIXED TAU FLAG; LAST TF COEFF IS FIXED TAU
C IPLAT :: NON-ZERO PLATEAU FLAG; READ LAST VALUE ON LINE3
C          IPLAT=1, fixed_plateau; IPLAT=2, variable_plateau
C IOLEV :: I/O LEVEL; 0=MINIMAL,1=FIT STATS,2=ITERATION DATA
C ICOL  :: COLUMN WITH OBSERVED DATA TO BE FIT; > 1, <= MXCOL
C
C CF    :: INITIAL PRE-EXP FIT COEFFICIENTS, ALWAYS NEXP-1 VALUES
C PLAT  :: OPTIONAL PLATEAU VALUE; MUST BE LAST, FOLLOWING CF VALUES
C TF    :: INITIAL TAU FIT COEFFICIENTS, ALWAYS NEXP VALUES

      if (iargc().ge.1) then
        call getarg(1,tmp)
        open(3,file=tmp,form='formatted',status='old')
      else
        goto 770
      endif
      c0 = 1.d0
C OPTION TO SET (REDUCE) THE FIRST POINT
      if (iargc().eq.2) then
        call getarg(2,tmp)
        read(tmp(1:10),'(F10.5)') c0
      endif
      read(3,'(A80)',end=770,err=770) fname
      read(3,*,end=770,err=770) nexp, ifixt,iplat,iolev,icol
      if (nexp.lt.1 .or. nexp.gt.4) goto 700
      if (icol.gt.mxcol) goto 710
      if (ifixt.ne.0 .and. ifixt.ne.1) goto 720
      if (iplat.ne.0 .and. iplat.ne.1 .and. iplat.ne.2) goto 730
      if (ifixt.eq.1 .and. iplat.eq.2) goto 735
      if (nexp.lt.2 .and. ifixt.eq.1) goto 740 
      if (nexp.eq.4 .and. iplat.eq.2) goto 745
      npre = nexp-1
      if (iplat.eq.0) then
        plat = zero
        read(3,*,end=770,err=770) (cf(j),j=1,npre)
      elseif (iplat.eq.1) then
        read(3,*,end=770,err=770) (cf(j),j=1,npre), plat
      elseif (iplat.eq.2) then
        plat = zero
        npre = npre + 1
        read(3,*,end=770,err=770) (cf(j),j=1,npre)
      endif
      ntau = nexp
      if (ifixt.eq.1) then
        ntau = ntau - 1
        read(3,*,end=770,err=770) (cf(j),j=npre+1,npre+ntau), tau
      elseif (ifixt.eq.0) then
        read(3,*,end=770,err=770) (cf(j),j=npre+1,npre+ntau)
      endif
      ncf = npre + ntau
      str = ' '
      read(3,'(A80)',end=20,err=20) str
   20 continue

C CHECK PRE-EXP CRITERIA; NON-NEG, SUM < ONE; SET PRE-EXP BOUNDS
      a = plat
      do i = 1,npre
        cfi(i) = cf(i)
        if(cf(i).le.zero) goto 750
        bnds(1,i) = tolr
        bnds(2,i) = c0 - tolr
        a = a + cf(i)
      enddo
      if (a.gt.c0) goto 760

C SET LOWER, UPPER BOUNDS FOR TAU BASED ON THE INITIAL VALS; NON-NEG TEST
      do i= npre+1,ncf
        cfi(i) = cf(i)
        if(cf(i).le.zero) goto 750
        bnds(1,i) =   0.01 * cf(i)
        bnds(2,i) = 100.00 * cf(i)
      enddo

C READ AND EXTRACT THE DATA
      open (9,file=fname,form='formatted',status='old')
      npts = 1
      do j = 1, mxpts
        read(9,*,end=100) (pts(k),k=1,icol)
        z(npts)  = pts(1)
        fz(npts) = pts(icol)
        npts = npts + 1
      enddo
      write(6,*) 'N.B. reached MXPTS before EOF at ',mxpts
 100  npts = npts - 1
      write(6,*) 'Read ',npts,' points from input'

C VERIFY (PRINT) INPUT PARAMETERS AND DATA
      write (6,*) 'Input data from ',fname(1:index(fname,' '))
      if (iplat.eq.1) write (6,*) 'Plateau enabled (last coeff.)'
      if (ifixt.eq.1) write (6,*) 'Fixed TAU enabled (last TAU coeff.)'
      write(6,901) npts, nexp, ncf, ntau
      write(6,903) (z(i),fz(i),i=1,2), (z(j),fz(j),j=npts-1,npts)
 901  format ('No. Obs. = ',I4,'  No. EXP = ',I1,'  Total coeff. = ',I1,
     *  '  No. TAU coeff. = ',I1)
 903  format (/,'Observed data endpoints:',/,4(2F10.4,/))

C INITIALIZE RANDOM NUMBER ROUTINE TO DEFAULT SEED
      call ran1
      nrept = 0

C RE-ENTRY POINT WITH NEW COEFF; SETUP ROUTINE
 101  continue
C SETUP ROUTINE; OUTPUT FLAGS SET BASED ON IOLEV, AFTER INITIALIZING
      call divset(1, iv, liv, lwk, wrk)
C SET OUTPUT OPTIONS BASED ON IOLEV  
      if (iolev.gt.1) iv(outlev) = -1
      if (iolev.gt.0) then
        iv(covprt) = 2
        iv(rdreq) = 2
        iv(statpr) = 1
        iv(solprt) = 1
      endif
C INCREASE ITERATION LIMITS mxfcal AND mxiter FOR REPEATS > ONE
      if (nrept.gt.1) then
        iv(17) = 500
        iv(18) = 200
      endif

C CALL NL FIT WITH DIFFERENT RESIDUAL ROUTINES
C CALL RESIDUAL ROUTINE AGAIN TO GET RESIDUALS FOR RMSRES
      k = 1
      if (ifixt.eq.0 .and. iplat.ne.2) then
C STRAIGHT EXPONENTIAL FIT OR WITH FIXED PLATEAU
        if (nexp.eq.1) then 
          call dnlafb(npts,ncf,cf,bnds,exp1f,iv,liv,lwk,wrk)
          call exp1f(npts,ncf,cf,k,r)
        elseif (nexp.eq.2) then 
          call dnlafb(npts,ncf,cf,bnds,exp2f,iv,liv,lwk,wrk)
          call exp2f(npts,ncf,cf,k,r)
        elseif (nexp.eq.3) then 
          call dnlafb(npts,ncf,cf,bnds,exp3f,iv,liv,lwk,wrk)
          call exp3f(npts,ncf,cf,k,r)
        elseif (nexp.eq.4) then 
          call dnlafb(npts,ncf,cf,bnds,exp4f,iv,liv,lwk,wrk)
          call exp4f(npts,ncf,cf,k,r)
        endif
      elseif (ifixt.eq.1 .and. iplat.ne.2) then 
C EXPONENTIAL FIT WITH A FIXED TAU VALUE, POSSIBLE FIXED PLATEAU
        if (nexp.eq.3) then 
          call dnlafb(npts,ncf,cf,bnds,exp3ft,iv,liv,lwk,wrk)
          call exp3ft(npts,ncf,cf,k,r)
        elseif (nexp.eq.4) then 
          call dnlafb(npts,ncf,cf,bnds,exp4ft,iv,liv,lwk,wrk)
          call exp4ft(npts,ncf,cf,k,r)
        elseif (nexp.eq.2) then 
          call dnlafb(npts,ncf,cf,bnds,exp2ft,iv,liv,lwk,wrk)
          call exp2ft(npts,ncf,cf,k,r)
        endif
      elseif (ifixt.eq.0 .and. iplat.eq.2) then 
C EXPONENTIAL FITS WITH VARIABLE PLATEAU
        if (nexp.eq.1) then 
          call dnlafb(npts,ncf,cf,bnds,exp1fp,iv,liv,lwk,wrk)
          call exp1fp(npts,ncf,cf,k,r)
        elseif (nexp.eq.2) then 
          call dnlafb(npts,ncf,cf,bnds,exp2fp,iv,liv,lwk,wrk)
          call exp2fp(npts,ncf,cf,k,r)
        elseif (nexp.eq.3) then 
          call dnlafb(npts,ncf,cf,bnds,exp3fp,iv,liv,lwk,wrk)
          call exp3fp(npts,ncf,cf,k,r)
        endif
      endif

C CHECK ERROR CODE, THEN CHECK FOR COEFFS "PEGGED" AT BOUNDARY
      call iv1err(iv(1),qrept)
      if ((iv(1).ge.3 .and. iv(1).le.6) .and. (.not. qrept)) then
        do i=1,ncf
          if(cf(i).eq.bnds(1,i) .or. cf(i).eq.bnds(2,i)) then
            qrept = .true.
          endif
        enddo
      endif

C RANDOMIZE COEFFS IF REPEAT FLAG IS SET
      if (qrept) then
        nrept = nrept + 1
C PREVENT EXCESSIVE ITERATIONS OF THIS
        if (nrept.gt.100) then
          write(6,*) 'Too many re-tries, giving up ...'
          stop 'the madness ...'
        endif
 121    continue
        a = zero
C DRANU() RETURNS A UNIFORM RANDOM REAL*8 NUMBER ON THE INTERVAL 0..1
        do i=1,npre
          cf(i) = 0.5*dranu()
          a = a + cf(i)
        enddo
        if (a.gt.c0-tolr) goto 121
        do i=npre+1,ncf
          cf(i) = cfi(i) + (dranu()-0.5d0)*cfi(i)
        enddo
        if(iolev.gt.0) then
          write(6,961) nrept,iv(1),(cf(i),i=1,npre)
 961      format(2I3,4F12.5)
          write(6,962) (cf(i),i=npre+1,ncf)
 962      format(6X,4F12.2)
        endif
        goto 101
      endif

      a = 0.d0
      do j=1,npts
        a = a + r(j)*r(j)
      enddo
      rmsr = sqrt(a/dble(npts))
      write(6,921) iv(1),rmsr
 921  format(' Exit code, IV(1)= ',I4,'  RMSres = ',F11.6,/)
      if(iplat.eq.1) write(6,'('' Fixed Plateau = '',F11.6)') plat
      k = ncf - ntau
      if(iplat.eq.2) write(6,*) ' * Variable plateau, CF', k
      write (6,*) 'Index  Final Coeff.  Initial       Bounds'
      write(6,922) (j,cf(j),cfi(j),bnds(1,j),bnds(2,j),j=1,k)
 922  format ('CF(',I1,')= ',4F11.6)
      write(6,923) ccf
 923  format ('CFCON= ',F11.6)
      write(6,924) (j,cf(j),cfi(j),bnds(1,j),bnds(2,j),j=k+1,ncf)
 924  format ('CF(',I1,')= ',4F11.3)
      if(ifixt.eq.1) write(6,'(''Fixed TAU = '',F11.3)') tau

C DERIVE OUTPUT FILE NAME, AND WRITE OBSERVED AND THEORETICAL DATA
      if (str(1:1).eq.' ') then
        j=index(fname,' ')
        do k=j-1,1,-1
          if (fname(k:k).eq.'.') goto 50
        enddo
   50   continue
        if (k.eq.1) then      
          str = fname(1:j-1)//'.fdt'
        else
          str = fname(1:k)//'fdt'
        endif
      endif
      write (6,*) 'Data written to ',str(1:index(str,' '))
      open(10,file=str,form='formatted')
      do i = 1,npts
        write(10,920) z(i),fz(i),tz(i)
      enddo
      close(10)
 920  format(F10.2,2F12.5)

C FLUSH STDOUT BEFORE PLOTTING
c     call oflush()

C FINALLY, PLOT THE FIT IF THE ENV VAR PGPLOT_DEV IS SET PROPERLY

      if (pgopen(' ') .gt. 0) then

      do i=1,npts
        time(i) = real(z(i))
        corr(i) = real(fz(i))
        theo(i) = real(tz(i))
      enddo

      call pgbbuf
      call pgsci(1)
c      call pgscf(2)
      call pgenv(0.,time(npts),0.,1.,0,0)
      write(tmp,931) icol, nexp, iplat,ifixt, npts
 931  format(1X,'Icol=',I2,'  Nexp=',I1,' Iplat=',I1,
     *  ' Ifixt=',I1,' Npts=',I5)
      ttl = fname(1:index(fname,' '))//tmp
      call pglab('Time (ps)','C(t)', ttl)

c      call pgscf(3)
      write(tmp,932) (cf(j),j=1,npre),ccf
 932  format ('pre-Exp = ',5F9.5)
      call pgptxt(.98*time(npts),0.96, 0.0,1.0,tmp(1:55))
      write(tmp,933) rmsr
 933  format ('RMS Res = ',E11.4)
      call pgptxt(0.0,1.01, 0.0,0.0,tmp(1:21))

      write(tmp,934) (cf(j), j=npre+1,ncf)
 934  format ('TAUs = ',4F9.2)
      do j=len(tmp),1,-1
        if (tmp(j:j).ne.' ') goto 250
      enddo
 250  continue
      if(ifixt.eq.1) write(tmp(j+3:j+18),935) tau
 935  format('FixTAU=',F9.2)
      call pgptxt(.98*time(npts),1.01, 0.0,1.0,tmp(1:j+18))
c      call pgscf(1)
      call pgebuf

      call pgbbuf
      call pgslw(2)
      call pgsci(2)
      call pgline(npts,time,corr)
      call pgslw(1)
      call pgsci(3)
      call pgline(npts,time,theo)
      call pgebuf

C SET VIEWPORT AND WINDOW FOR SUBWINDOW (PGBOX)
      call pgqvp(0,x1,x2,y1,y2)
      call pgsvp(0.45*(x2-x1),x2, (0.45*(y2-y1)),y2)
      k = npts/5
      x2 = time(k)
      c = pgrnd(x2/5.,i)
      y1 = corr(k)
      y2 = corr(4)
      y1 = y1 - .05*y1
      y2 = y2 + .05*y2
      call pgswin(0.,x2, y1, y2)
      call pgsci(1)
      call pgbbuf
      call pgbox('BCNT',c,1, 'BCNT',0.1,1)
C REDRAW THE DATA IN THE SUBWINDOW
      call pgslw(2)
      call pgsci(2)
      call pgline(npts,time,corr)
      call pgslw(1)
      call pgsci(3)
      call pgline(npts,time,theo)
      call pgebuf

      call pgclos

C PGPLOT BYPASS 
      else
        write(6,*) 'PGOPEN failed, check env var PGPLOT_DEV'
      endif
C PROGRAM END; BAD INPUT ERROR MESSAGES
      stop 'Normal Exit'

 690  write (6,*) 'Bad no. of pts, exceeds MXPTS:',npts
      stop 'ERROR'
 700  write (6,*) 'Bad no. of EXP, 1-4 only:',nexp
      stop 'ERROR'
 710  write (6,*) 'icol (',icol,') > mxcol (',mxcol,')'
      stop 'ERROR'
 720  write (6,*) 'Bad fixed TAU flag (0,1):',ifixt
      stop 'ERROR'
 730  write (6,*) 'Bad plateau flag (0,1,2):',iplat
      stop 'ERROR'
 735  write (6,*) 'Fixed TAU, variable plateau',ifixt,iplat
      stop 'ERROR'
 740  write (6,*) 'Fixed TAU with nEXP < 2:',ifixt,nexp
      stop 'ERROR'
 745  write (6,*) 'Variable plateau, nEXP = 4:',iplat,nexp
      stop 'ERROR'
 750  write (6,*) 'Bad coeff., < zero:', cf(i)
      stop 'ERROR'
 760  write (6,*) 'Bad coeffs., sum > C0:', a, (cf(j),j=1,k), plat
      stop 'ERROR'
 770  write (6,*) '  expnfit.x  Control_File  [ CZERO ]'
      write (6,*) ' '
      write (6,*) 'Control_File is a 4 (or 5) line file with either a'
      write (6,*) 'filename or number data (integer,real) on each line'
      write (6,*) ' '
      write (6,*) 'CZERO  Optional arg for sum of pre-Exp coeffs.'
      write (6,*) ' '
      write (6,*) '======== Control_File Structure & Contents ======'
      write (6,*) 'LINE1  Input_Data_File_Name (e.g. filename.ext)'
      write (6,*) '  [text, up to 12 values/line, no missing data]'
      write (6,*) 'LINE2  NEXP IFIXT IPLAT IOLEV ICOL'
      write (6,*) 'LINE3  (CF(J),J=1,NEXP-1), [ PLAT ]'
      write (6,*) 'LINE4  (TF(J),J=1,NEXP)'
      write (6,*) 'LINE5  Output_Data_File_Name (optional)'
      write (6,*) '  [name derived from Input if not supplied]'
      write (6,*) ' ------------ LINE1, LINE5 --------------------'
      write (6,*) ' text string with no leading or embedded blanks'
      write (6,*) ' ---------- LINE2, INTEGER --------------------'
      write (6,*) 'NEXP   No. of exponentials to use for fit'
      write (6,*) 'IFIXT  Fixed TAU flag; last TF coeff is fixed TAU'
      write (6,*) 'IPLAT  Plateau flag; 0=None, 1=Fixed, 2=Variable'
      write (6,*) 'IOLEV  0=Minimal, 1=Fit stats, 2=Iteration data'
      write (6,*) 'ICOL   Index of column with obs. data; icol LE 12'
      write (6,*) ' ------------- LINE3, REAL --------------------'
      write (6,*) 'CF     Initial pre-Exp coeff., NEXP-1 values'
      write (6,*) 'PLAT   Optional plateau; must follow CF vals'
      write (6,*) ' ------------- LINE4, REAL --------------------'
      write (6,*) 'TF     Initial TAU coeff., NEXP values'
      write (6,*) '================================================='
      write (6,*) ' '
      write (6,*) 'EXAMPLE: 3 exponential fit to col 4 of corrfxn.dat,'
      write (6,*) '  [prog assumes time in col 1]'
      write (6,*) ' '
      write (6,*) 'corrfxn.dat'
      write (6,*) ' 3 0 0 2 4'
      write (6,*) ' 0.4  0.3'
      write (6,*) ' 2.0 40. 500.'
      write (6,*) ' '
      stop 'Bad Control_File Exit'
      end

C ERROR ANALYSIS ROUTINE; RETURNS TRUE FOR A RETRY, TRAP FATAL ERRORS
      subroutine iv1err(iv1,qerr)
      logical qerr
      integer iv1
      if (iv1.ge.3 .and. iv1.le.6) then
        qerr = .false.
        return
      elseif (iv1.ge.7 .and. iv1.le.13) then
        qerr = .true.
        return
      elseif (iv1.eq.14) then
        qerr = .false.
        return
      elseif (iv1.eq.63 .or. iv1.eq.65) then
        qerr = .true.
        return
      else
        write(6,*) 'Oops, I can''t go on with IV(1) = ',iv1
        stop 'ERROR EXIT'
      endif
      return
      end


C THE RESIDUAL ROUTINES; FIT TO WEIGHTED SUM OF EXPONENTIALS
C FOR N EXP THERE ARE N-1 WEIGHTS; MUST SUM TO ONE (CONSTRAINT)
C NOTE THE PLATEAU IS ALWAYS PRESENT, BUT IS ZERO BY DEFAULT

      subroutine exp1f(no, nc, cf, icnt, res)
      integer no, nc, icnt, i, mxpts
      real*8 cf(nc), res(no)
      parameter (mxpts=500000)
      real*8  z(mxpts),fz(mxpts), tz(mxpts), plat,tau,ccf,c0
      common /fitdat/ z, fz, tz, plat, tau, ccf, c0
      ccf = c0 - plat
      if(ccf.lt.0.d0) then
        icnt = 0
        return
      endif
      do i=1,no
        tz(i) = plat + ccf*exp(-z(i)/cf(1))
        res(i) = fz(i) - tz(i)
      enddo
      return
      end

      subroutine exp2f(no, nc, cf, icnt, res)
      integer no, nc, icnt, i, mxpts
      real*8 cf(nc), res(no)
      parameter (mxpts=500000)
      real*8  z(mxpts),fz(mxpts), tz(mxpts), plat,tau,ccf,c0
      common /fitdat/ z, fz, tz, plat, tau, ccf, c0
      ccf = c0 - (cf(1)+plat)
      if(ccf.lt.0.d0) then
        icnt = 0
        return
      endif
      do i=1,no
        tz(i) = plat + cf(1)*exp(-z(i)/cf(2)) + ccf*exp(-z(i)/cf(3))
        res(i) = fz(i) - tz(i)
      enddo
      return
      end

      subroutine exp3f(no, nc, cf, icnt, res)
      integer no, nc, icnt, i, mxpts
      real*8 cf(nc), res(no)
      parameter (mxpts=500000)
      real*8  z(mxpts),fz(mxpts), tz(mxpts), plat,tau,ccf,c0
      common /fitdat/ z, fz, tz, plat, tau, ccf, c0
      ccf = c0 - (cf(1)+cf(2)+plat)
      if(ccf.lt.0.d0) then
        icnt = 0
        return
      endif
      do i=1,no
        tz(i) = plat + cf(1)*exp(-z(i)/cf(3)) + cf(2)*exp(-z(i)/cf(4))
     *    + ccf*exp(-z(i)/cf(5))
        res(i) = fz(i) - tz(i)
      enddo
      return
      end

      subroutine exp4f(no, nc, cf, icnt, res)
      integer no, nc, icnt, i, mxpts
      real*8 cf(nc), res(no)
      parameter (mxpts=500000)
      real*8  z(mxpts),fz(mxpts), tz(mxpts), plat,tau,ccf,c0
      common /fitdat/ z, fz, tz, plat, tau, ccf, c0
      ccf = c0 - (cf(1)+cf(2)+cf(3)+plat)
      if(ccf.lt.0.d0) then
        icnt = 0
        return
      endif
      do i=1,no
        tz(i) = plat + cf(1)*exp(-z(i)/cf(4)) + cf(2)*exp(-z(i)/cf(5))
     *    + cf(3)*exp(-z(i)/cf(6)) + ccf*exp(-z(i)/cf(7))
        res(i) = fz(i) - tz(i)
      enddo
      return
      end

C VARIANTS FOR 2, 3 AND 4 EXP WITH ONE TAU FIXED

      subroutine exp2ft(no, nc, cf, icnt, res)
      integer no, nc, icnt, i, mxpts
      real*8 cf(nc), res(no)
      parameter (mxpts=500000)
      real*8  z(mxpts),fz(mxpts), tz(mxpts), plat,tau,ccf,c0
      common /fitdat/ z, fz, tz, plat, tau, ccf, c0
      ccf = c0 - (cf(1)+plat)
      if(ccf.lt.0.d0) then
        icnt = 0
        return
      endif
      do i=1,no
        tz(i) = plat + cf(1)*exp(-z(i)/cf(2)) + ccf*exp(-z(i)/tau)
        res(i) = fz(i) - tz(i)
      enddo
      return
      end

      subroutine exp3ft(no, nc, cf, icnt, res)
      integer no, nc, icnt, i, mxpts
      real*8 cf(nc), res(no)
      parameter (mxpts=500000)
      real*8  z(mxpts),fz(mxpts), tz(mxpts), plat,tau,ccf,c0
      common /fitdat/ z, fz, tz, plat, tau, ccf, c0
      ccf = c0 - (cf(1)+cf(2)+plat)
      if(ccf.lt.0.d0) then
        icnt = 0
        return
      endif
      do i=1,no
        tz(i) = plat + cf(1)*exp(-z(i)/cf(3)) + cf(2)*exp(-z(i)/cf(4))
     *    + ccf*exp(-z(i)/tau)
        res(i) = fz(i) - tz(i)
      enddo
      return
      end

      subroutine exp4ft(no, nc, cf, icnt, res)
      integer no, nc, icnt, i, mxpts
      real*8 cf(nc), res(no)
      parameter (mxpts=500000)
      real*8  z(mxpts),fz(mxpts), tz(mxpts), plat,tau,ccf,c0
      common /fitdat/ z, fz, tz, plat, tau, ccf, c0
      ccf = c0 - (cf(1)+cf(2)+cf(3)+plat)
      if(ccf.lt.0.d0) then
        icnt = 0
        return
      endif
      do i=1,no
        tz(i) = plat + cf(1)*exp(-z(i)/cf(4)) + cf(2)*exp(-z(i)/cf(5))
     *    + cf(3)*exp(-z(i)/cf(6)) + ccf*exp(-z(i)/tau)
        res(i) = fz(i) - tz(i)
      enddo
      return
      end

C VARIABLE PLATEAU FOR ONE, TWO, AND THREE EXPONENTIALS

      subroutine exp1fp(no, nc, cf, icnt, res)
      integer no, nc, icnt, i, mxpts
      real*8 cf(nc), res(no)
      parameter (mxpts=500000)
      real*8  z(mxpts),fz(mxpts), tz(mxpts), plat,tau,ccf,c0
      common /fitdat/ z, fz, tz, plat, tau, ccf, c0
      ccf = c0 - cf(1)
      if(ccf.lt.0.d0) then
        icnt = 0
        return
      endif
      do i=1,no
        tz(i) = cf(1) + ccf*exp(-z(i)/cf(2))
        res(i) = fz(i) - tz(i)
      enddo
      return
      end

      subroutine exp2fp(no, nc, cf, icnt, res)
      integer no, nc, icnt, i, mxpts
      real*8 cf(nc), res(no)
      parameter (mxpts=500000)
      real*8  z(mxpts),fz(mxpts), tz(mxpts), plat,tau,ccf,c0
      common /fitdat/ z, fz, tz, plat, tau, ccf, c0
      ccf = c0 - (cf(1)+cf(2))
      if(ccf.lt.0.d0) then
        icnt = 0
        return
      endif
      do i=1,no
        tz(i) = cf(2) + cf(1)*exp(-z(i)/cf(3)) + ccf*exp(-z(i)/cf(4))
        res(i) = fz(i) - tz(i)
      enddo
      return
      end

      subroutine exp3fp(no, nc, cf, icnt, res)
      integer no, nc, icnt, i, mxpts
      real*8 cf(nc), res(no)
      parameter (mxpts=500000)
      real*8  z(mxpts),fz(mxpts), tz(mxpts), plat,tau,ccf,c0
      common /fitdat/ z, fz, tz, plat, tau, ccf, c0
      ccf = c0 - (cf(1)+cf(2)+cf(3))
      if(ccf.lt.0.d0) then
        icnt = 0
        return
      endif
      do i=1,no
        tz(i) = cf(3) + cf(1)*exp(-z(i)/cf(4)) + cf(2)*exp(-z(i)/cf(5))
     *    + ccf*exp(-z(i)/cf(6))
        res(i) = fz(i) - tz(i)
      enddo
      return
      end

